from typing import Any, Dict, Optional

# noinspection PyPackageRequirements
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkParamTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    def __init__(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super(FrameworkParamTransformer, self).__init__()

        self.progress_logger: Optional[ProgressLogger] = progress_logger
        self.parameters: Optional[Dict[str, Any]] = parameters

    def _transform(self, df: DataFrame, response: Dict[str, Any]) -> Any:  # type: ignore
        return response

    def transform(  # type: ignore
        self, dataset: DataFrame, response: Dict[str, Any], params=None
    ) -> Any:
        if params is None:
            params = dict()
        if isinstance(params, dict):
            if params:
                return self.copy(params)._transform(dataset, response)
            else:
                return self._transform(dataset, response)
        else:
            raise ValueError("Params must be a param map but got %s." % type(params))



from typing import Dict, Any, Optional, cast

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkJdbcReader(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        jdbc_url: str,
        query: str,
        driver: str,
        view: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        assert jdbc_url
        assert query
        assert driver

        self.logger = get_logger(__name__)

        # add a param
        self.jdbc_url: Param[str] = Param(self, "jdbc_url", "")
        self._setDefault(jdbc_url=jdbc_url)

        self.query: Param[str] = Param(self, "query", "")
        self._setDefault(query=query)

        self.driver: Param[str] = Param(self, "driver", "")
        self._setDefault(driver=driver)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        # pushdown_query = "(select * from employees where emp_no < 10008) emp_alias"
        query: str = self.getQuery()
        jdbc_url: str = self.getJdbcUrl()
        driver: str = self.getDriver()
        view: Optional[str] = self.getView()
        df = (
            # this execution requires an Option either 'dbtable' or 'query' parameter
            df.sql_ctx.read.format("jdbc")
            .option("url", jdbc_url)
            .option("dbtable", query)
            .option("driver", driver)
            .load()
        )

        if view:
            df.createOrReplaceTempView(view)

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getJdbcUrl(self) -> str:
        return cast(str, self.getOrDefault(self.jdbc_url))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDriver(self) -> str:
        return cast(str, self.getOrDefault(self.driver))

    # noinspection PyPep8Naming,PyMethodMayBeStatic
    def getFormat(self) -> str:
        return "jdbc"

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getQuery(self) -> str:
        return cast(str, self.getOrDefault(self.query))

    # noinspection PyPep8Naming
    def getView(self) -> Optional[str]:
        return cast(Optional[str], self.getOrDefault(self.view))



import shutil
from os import path
from pathlib import Path
from typing import Dict, Any

import pytest
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType
from tests.conftest import clean_spark_session

from library.features.carriers_python.v1.features_carriers_python_v1 import (
    FeaturesCarriersPythonV1,
)

from library.features.carriers.v1.features_carriers_v1 import FeaturesCarriersV1

from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import (
    FrameworkCsvLoader,
)

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.pipelines.v2.framework_pipeline import FrameworkPipeline
from spark_pipeline_framework.transformers.framework_validation_transformer.v1.framework_validation_transformer import (
    FrameworkValidationTransformer,
)


class MyUnValidatedPipeline(FrameworkPipeline):
    def __init__(self, parameters: Dict[str, Any], progress_logger: ProgressLogger):
        super(MyUnValidatedPipeline, self).__init__(
            parameters=parameters, progress_logger=progress_logger, run_id="12345678"
        )
        self.transformers = self.create_steps(
            [  # type: ignore
                FrameworkCsvLoader(
                    view="flights",
                    filepath=parameters["flights_path"],
                    parameters=parameters,
                    progress_logger=progress_logger,
                ),
                FeaturesCarriersV1(
                    parameters=parameters, progress_logger=progress_logger
                ),
                FeaturesCarriersPythonV1(
                    parameters=parameters, progress_logger=progress_logger
                ),
            ]
        )


class MyValidatedPipeline(FrameworkPipeline):
    def __init__(self, parameters: Dict[str, Any], progress_logger: ProgressLogger):
        super(MyValidatedPipeline, self).__init__(
            parameters=parameters,
            progress_logger=progress_logger,
            run_id="12345678",
            validation_output_path=parameters["validation_output_path"],
        )
        self.transformers = self.create_steps(
            [  # type: ignore
                FrameworkCsvLoader(
                    view="flights",
                    filepath=parameters["flights_path"],
                    parameters=parameters,
                    progress_logger=progress_logger,
                ),
                FrameworkValidationTransformer(
                    validation_source_path=parameters["validation_source_path"],
                    validation_queries=["validate.sql"],
                ),
                FeaturesCarriersV1(
                    parameters=parameters, progress_logger=progress_logger
                ),
                FeaturesCarriersPythonV1(
                    parameters=parameters, progress_logger=progress_logger
                ),
            ]
        )


class MyFailFastValidatedPipeline(FrameworkPipeline):
    def __init__(self, parameters: Dict[str, Any], progress_logger: ProgressLogger):
        super(MyFailFastValidatedPipeline, self).__init__(
            parameters=parameters,
            progress_logger=progress_logger,
            run_id="12345678",
            validation_output_path=parameters["validation_output_path"],
        )
        self.transformers = self.create_steps(
            [  # type: ignore
                FrameworkCsvLoader(
                    view="flights",
                    filepath=parameters["flights_path"],
                    parameters=parameters,
                    progress_logger=progress_logger,
                ),
                FrameworkValidationTransformer(
                    validation_source_path=parameters["validation_source_path"],
                    validation_queries=["validate.sql"],
                    fail_on_validation=True,
                ),
                FeaturesCarriersV1(
                    parameters=parameters, progress_logger=progress_logger
                ),
                FeaturesCarriersPythonV1(
                    parameters=parameters, progress_logger=progress_logger
                ),
            ]
        )


def test_can_run_validated_framework_pipeline(spark_session: SparkSession) -> None:
    with pytest.raises(AssertionError):
        # Arrange
        clean_spark_session(spark_session)
        data_dir: Path = Path(__file__).parent.joinpath("./")
        flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"
        output_path: str = (
            f"file://{data_dir.joinpath('temp').joinpath('validation.csv')}"
        )

        if path.isdir(data_dir.joinpath("temp")):
            shutil.rmtree(data_dir.joinpath("temp"))

        schema = StructType([])

        df: DataFrame = spark_session.createDataFrame(
            spark_session.sparkContext.emptyRDD(), schema
        )

        spark_session.sql("DROP TABLE IF EXISTS default.flights")

        # Act
        parameters = {
            "flights_path": flights_path,
            "validation_source_path": str(data_dir),
            "validation_output_path": output_path,
        }

        with ProgressLogger() as progress_logger:
            pipeline: MyValidatedPipeline = MyValidatedPipeline(
                parameters=parameters, progress_logger=progress_logger
            )
            transformer = pipeline.fit(df)
            transformer.transform(df)


def test_can_run_unvalidated_framework_pipeline(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)
    data_dir: Path = Path(__file__).parent.joinpath("./")
    flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    spark_session.sql("DROP TABLE IF EXISTS default.flights")

    # Act
    parameters = {"flights_path": flights_path}

    with ProgressLogger() as progress_logger:
        pipeline: MyUnValidatedPipeline = MyUnValidatedPipeline(
            parameters=parameters, progress_logger=progress_logger
        )
        transformer = pipeline.fit(df)
        transformer.transform(df)

    # Assert
    result_df: DataFrame = spark_session.sql("SELECT * FROM flights2")
    result_df.show()

    assert result_df.count() > 0


def test_validated_framework_pipeline_writes_results(
    spark_session: SparkSession,
) -> None:
    # Arrange
    clean_spark_session(spark_session)
    data_dir: Path = Path(__file__).parent.joinpath("./")
    flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"
    output_path: str = f"file://{data_dir.joinpath('temp').joinpath('validation.csv')}"

    if path.isdir(data_dir.joinpath("temp")):
        shutil.rmtree(data_dir.joinpath("temp"))

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    spark_session.sql("DROP TABLE IF EXISTS default.flights")

    # Act
    parameters = {
        "flights_path": flights_path,
        "validation_source_path": str(data_dir),
        "validation_output_path": output_path,
    }

    try:
        with ProgressLogger() as progress_logger:
            pipeline: MyValidatedPipeline = MyValidatedPipeline(
                parameters=parameters, progress_logger=progress_logger
            )
            transformer = pipeline.fit(df)
            transformer.transform(df)
    except AssertionError:
        validation_df = df.sql_ctx.read.csv(output_path, header=True)
        validation_df.show(truncate=False)
        assert validation_df.count() == 1


def test_fail_fast_validated_framework_pipeline_writes_results(
    spark_session: SparkSession,
) -> None:
    # Arrange
    clean_spark_session(spark_session)
    data_dir: Path = Path(__file__).parent.joinpath("./")
    flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"
    output_path: str = f"file://{data_dir.joinpath('temp').joinpath('validation.csv')}"

    if path.isdir(data_dir.joinpath("temp")):
        shutil.rmtree(data_dir.joinpath("temp"))

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    spark_session.sql("DROP TABLE IF EXISTS default.flights")

    # Act
    parameters = {
        "flights_path": flights_path,
        "validation_source_path": str(data_dir),
        "validation_output_path": output_path,
    }

    try:
        with ProgressLogger() as progress_logger:
            pipeline: MyFailFastValidatedPipeline = MyFailFastValidatedPipeline(
                parameters=parameters, progress_logger=progress_logger
            )
            transformer = pipeline.fit(df)
            transformer.transform(df)
    except AssertionError:
        validation_df = df.sql_ctx.read.csv(output_path, header=True)
        validation_df.show(truncate=False)
        assert validation_df.count() == 1



from typing import Any, Dict, List, Optional, TYPE_CHECKING

# noinspection PyPackageRequirements
from pyspark.ml.base import Transformer

# noinspection PyPackageRequirements
from pyspark.ml.param import Param

# noinspection PyPackageRequirements
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable

# noinspection PyPackageRequirements
from pyspark.sql.dataframe import DataFrame
from typing_extensions import final

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger


class FrameworkTransformer(
    Transformer,
    DefaultParamsReadable,  # type: ignore
    DefaultParamsWritable,
):
    # noinspection PyUnusedLocal
    def __init__(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super(FrameworkTransformer, self).__init__()

        if TYPE_CHECKING:
            self._input_kwargs: Dict[str, Any] = {}
        self.logger = get_logger(__name__)

        self.name: Param[str] = Param(self, "name", "")
        self._setDefault(name=name)

        self.progress_logger: Optional[ProgressLogger] = progress_logger

        self.parameters: Optional[Dict[str, Any]] = parameters

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal
    def setStandardParams(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> Any:
        kwargs: Dict[str, Any] = self._input_kwargs
        return self._set(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring, PyUnusedLocal,Mypy
    @final
    def setParams(self, **kwargs: Any) -> Any:
        # ignore any parameters
        kwargs = {key: value for key, value in kwargs.items() if self.hasParam(key)}
        return self._set(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        return df

    @property
    def transformers(self) -> List[Transformer]:
        return [self]

    # # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    # def setName(self, value: str) -> "FrameworkTransformer":
    #     self._paramMap[self.name] = value
    #     return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name)  # type: ignore

    # # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    # def setProgressLogger(
    #     self, value: Optional[ProgressLogger]
    # ) -> "FrameworkTransformer":
    #     self._paramMap[self.progress_logger] = value
    #     return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getProgressLogger(self) -> Optional[ProgressLogger]:
        return self.progress_logger

    # # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    # def setParameters(self, value: Optional[Dict[str, Any]]) -> "FrameworkTransformer":
    #     self._paramMap[self.parameters] = value
    #     return self
    #
    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getParameters(self) -> Optional[Dict[str, Any]]:
        return self.parameters

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSql(self) -> Optional[str]:
        return None

    # # This is here to avoid mypy from complaining since this is a protected member
    # noinspection PyPep8Naming
    def _setDefault(self, **kwargs: Any) -> None:
        # noinspection PyUnresolvedReferences,PyProtectedMember
        super()._setDefault(**kwargs)

    def _set(self, **kwargs: Any) -> None:
        # filter out any args that don't have parameters
        kwargs = {key: value for key, value in kwargs.items() if self.hasParam(key)}
        # noinspection PyUnresolvedReferences,PyProtectedMember
        super()._set(**kwargs)

    def __str__(self) -> str:
        stage_name: str = ""
        stage_name += f"name={self.getName()} "
        if hasattr(self, "getView"):
            # noinspection Mypy
            stage_name += f"view={self.getView()} "
        stage_name += "type=" + self.__class__.__name__
        return stage_name



from typing import List, Any


def flatten(my_list: List[Any]) -> List[Any]:
    """

    :param my_list:
    :return:
    """
    if not my_list:
        return my_list
    flat_list: List[Any] = []
    for element in my_list:
        if isinstance(element, list):
            flat_list.extend(flatten(element))
        else:
            flat_list.append(element)
    return flat_list


# def create_stages(transformer_list: List, parameters, progress_logger):
#     return [t(parameters, progress_logger) for t in flatten(transformer_list)]



import logging
import os
import shutil

# noinspection PyPackageRequirements
from typing import Any

import pytest
from pyspark.sql import SparkSession

# make sure env variables are set correctly
if "SPARK_HOME" not in os.environ:
    os.environ["SPARK_HOME"] = "/usr/local/opt/spark"


def quiet_py4j() -> None:
    """turn down spark logging for the carriers context"""
    logger = logging.getLogger("py4j")
    logger.setLevel(logging.ERROR)


def clean_spark_dir() -> None:
    """

    :return:
    """
    try:
        os.remove("./derby.log")
        shutil.rmtree("./metastore_db")
        shutil.rmtree("./spark-warehouse")
    except OSError:
        pass


def clean_spark_session(session: SparkSession) -> None:
    """

    :param session:
    :return:
    """
    tables = session.catalog.listTables("default")

    for table in tables:
        print(f"clear_tables() is dropping table/view: {table.name}")
        # noinspection SqlDialectInspection,SqlNoDataSourceInspection
        session.sql(f"DROP TABLE IF EXISTS default.{table.name}")
        # noinspection SqlDialectInspection,SqlNoDataSourceInspection
        session.sql(f"DROP VIEW IF EXISTS default.{table.name}")
        # noinspection SqlDialectInspection,SqlNoDataSourceInspection
        session.sql(f"DROP VIEW IF EXISTS {table.name}")

    session.catalog.clearCache()


def clean_close(session: SparkSession) -> None:
    """

    :param session:
    :return:
    """
    clean_spark_session(session)
    clean_spark_dir()
    session.stop()


@pytest.fixture(scope="session")
def spark_session(request: Any) -> SparkSession:
    # make sure env variables are set correctly
    if "SPARK_HOME" not in os.environ:
        os.environ["SPARK_HOME"] = "/usr/local/opt/spark"

    clean_spark_dir()

    master = "local[2]"

    session = (
        SparkSession.builder.appName("pytest-pyspark-local-testing")
        .master(master)
        .config("spark.ui.showConsoleProgress", "false")
        .config("spark.sql.shuffle.partitions", "2")
        .config("spark.default.parallelism", "4")
        .config("spark.sql.broadcastTimeout", "2400")
        .config(
            "spark.jars.packages",
            "com.crealytics:spark-excel_2.12:0.13.7,com.databricks:spark-xml_2.12:0.14.0",
        )
        .enableHiveSupport()
        .getOrCreate()
    )

    request.addfinalizer(lambda: clean_close(session))
    quiet_py4j()
    return session



from typing import Any, Dict, Optional

from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_param_transformer.v1.framework_param_transformer import (
    FrameworkParamTransformer,
)
from spark_pipeline_framework.utilities.scrapy_custom_crawler import ScrapyCustomCrawler  # type: ignore


class FrameworkWebCrawler(FrameworkParamTransformer):
    def __init__(
        self,
        spider_class: type,
        name: str = None,  # type: ignore
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        response_key: Optional[str] = "urls",
    ) -> None:
        """
        Framework to begin the crawl process for the defined spiders

        :param spider_class: type <Spider class>: Spider class (spider) which has to crawl
        :param response_key: str [optional]: Parameters key which will hold the response from the crawler
        """
        assert name

        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        self.spider_class = spider_class
        self.response_key = response_key or "urls"

    def _transform(self, df: DataFrame, response: Dict[str, Any]) -> Dict[str, Any]:  # type: ignore
        crawler = ScrapyCustomCrawler()
        crawler.crawl(self.spider_class)
        out = crawler.output.get("response", [])
        response[self.response_key] = out
        return response



from typing import Optional

from pyspark.sql import SparkSession


def save_csv_as_parquet(
    spark_session: SparkSession,
    source_path: str,
    destination_path: str,
    delimiter: str = ",",
    has_header: bool = True,
    infer_schema: bool = True,
    num_partitions: Optional[int] = None,
) -> None:
    """
    Saves csv as parquet

    """
    # noinspection SpellCheckingInspection
    spark_session.read.option("delimiter", delimiter).option(
        "header", str(has_header)
    ).option("inferschema", infer_schema).csv(source_path).repartition(
        num_partitions or 1
    ).write.format(
        "parquet"
    ).mode(
        "overwrite"
    ).save(
        destination_path
    )



from logging import Logger
from pathlib import Path
from typing import Union, List, Optional, Dict, Any, cast

from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql import DataFrameReader
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import StructType

from spark_pipeline_framework.logger.yarn_logger import get_logger

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.file_helpers import get_absolute_paths


class FrameworkXmlLoader(FrameworkTransformer):
    """
    Load xml files into a DataFrame by specifying a path and an optional schema
    """

    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        view: str,
        filepath: Union[str, List[str], Path],
        row_tag: str,
        schema: Optional[StructType] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> None:
        """
        Initializes the framework_xml_loader
        :param view: The name of the view that the resultant DataFrame will be stored in
        :param filepath: The path to the xml file to load
        :param row_tag: The row tag of your xml files to treat as a row
        :param schema: The schema to apply to the DataFrame, if not passed schema will be inferred
        :param name: sets the name of the transformer as it will appear in logs
        :param parameters:
        :param progress_logger:
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        self.logger: Logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=None)

        self.filepath: Param[str] = Param(self, "filepath", "")
        self._setDefault(filepath=None)

        self.row_tag: Param[str] = Param(self, "row_tag", "")
        self._setDefault(row_tag=None)

        self.schema: Param[StructType] = Param(self, "schema", "")
        self._setDefault(schema=None)

        if not filepath:
            raise ValueError("filepath is None or empty")

        self.logger.info(f"Received filepath: {filepath}")

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view = self.getView()
        filepath: Union[str, List[str], Path] = self.getFilepath()
        row_tag: str = self.getRowTag()
        schema: StructType = self.getSchema()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        paths = get_absolute_paths(filepath=filepath)
        progress_logger and progress_logger.write_to_log(
            f"Loading file for view {view}: {paths}"
        )

        df_xml_reader: DataFrameReader = df.sql_ctx.read.format("xml").options(
            rowTag=row_tag
        )
        if schema:
            df_xml = df_xml_reader.load(paths, schema=schema)
        else:
            df_xml = df_xml_reader.load(paths)
        df_xml.createOrReplaceTempView(view)
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return cast(str, self.getOrDefault(self.view))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilepath(self) -> Union[str, List[str], Path]:
        return cast(Union[str, List[str], Path], self.getOrDefault(self.filepath))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getRowTag(self) -> str:
        return cast(str, self.getOrDefault(self.row_tag))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSchema(self) -> StructType:
        return cast(StructType, self.getOrDefault(self.schema))



from typing import Any, List

from spark_pipeline_framework.utilities.flattener import flatten


def test_flatten_simple_list() -> None:
    my_list: List[Any] = [1, 2, 3]

    assert my_list == flatten(my_list=my_list)


def test_flatten_nested_list() -> None:
    my_list: List[Any] = [[1], [2, 3], [4]]

    assert [1, 2, 3, 4] == flatten(my_list=my_list)


def test_flatten_nested_two_levels_list() -> None:
    my_list: List[Any] = [[1], [[2, 3]], [4]]

    assert [1, 2, 3, 4] == flatten(my_list=my_list)



from typing import Dict, Any, Optional, List, cast

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkDropDuplicatesTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        view: str,
        columns: List[str],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        assert view
        assert columns
        assert isinstance(columns, list)

        # add a param
        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.columns: Param[List[str]] = Param(self, "columns", "")
        self._setDefault(columns=columns)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        columns: List[str] = self.getColumns()

        result_df: DataFrame = df.sql_ctx.table(view)
        result_df = result_df.drop_duplicates(columns)
        result_df.createOrReplaceTempView(view)

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return cast(str, self.getOrDefault(self.view))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getColumns(self) -> List[str]:
        return cast(List[str], self.getOrDefault(self.columns))



import os
from pathlib import Path
from typing import Dict, Any

from library.pipelines.reference_data_download_test_pipeline.v1.test_spider_class.python_wget_module_test_spider import (
    PythonWgetModuleTestSpider,
)
from spark_pipeline_framework.pipelines.framework_param_pipeline import (
    FrameworkParamPipeline,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.transformers.framework_file_downloader.v1.framework_file_downloader import (
    FrameworkFileDownloader,
)
from spark_pipeline_framework.transformers.framework_web_crawler.v1.framework_web_crawler import (
    FrameworkWebCrawler,
)


class ReferenceDataDownloadTestPipeline(FrameworkParamPipeline):
    def __init__(self, parameters: Dict[str, Any], progress_logger: ProgressLogger):
        super(ReferenceDataDownloadTestPipeline, self).__init__(
            parameters=parameters, progress_logger=progress_logger
        )

        download_path: str = os.path.join(Path(__file__).parent, "data")

        self.transformers = self.create_steps(
            [
                FrameworkWebCrawler(
                    spider_class=PythonWgetModuleTestSpider,
                    name="python_wget_module_test",
                    parameters=parameters,
                    progress_logger=progress_logger,
                ),
                FrameworkFileDownloader(
                    download_to_path=download_path,
                    extract_zips=True,
                    parameters=parameters,
                    progress_logger=progress_logger,
                ),
            ]
        )



from typing import Any, Dict

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from spark_pipeline_framework.transformers.framework_web_crawler.v1.framework_web_crawler import (
    FrameworkWebCrawler,
)
from tests.conftest import clean_spark_session
from tests.transformers.framework_web_crawler.v1.spider_classes.spider_test_class import (  # type: ignore
    SpiderTestClass,
)


def test_web_crawler(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)
    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    response: Dict[str, Any] = {}

    # Act
    response = FrameworkWebCrawler(
        spider_class=SpiderTestClass, name="test_crawler"
    ).transform(df, response)

    # Assert
    print(response)
    assert response



from typing import Optional, Dict, Any, List, cast

from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql import DataFrame

from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_param_transformer.v1.framework_param_transformer import (
    FrameworkParamTransformer,
)
from spark_pipeline_framework.utilities.file_downloader import FileDownloader


class FrameworkFileDownloader(FrameworkParamTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        download_urls: Optional[List[str]] = [],
        download_to_path: Optional[str] = "./",
        param_key: Optional[str] = "urls",
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        extract_zips: Optional[bool] = False,
    ) -> None:
        """
        Framework to download files from urls

        :param download_urls: list(str) [optional]: List of urls from which the files have to be downloaded
        :param download_to_path: str [optional]: Path where the file has to be downloaded
        :param param_key: str [optional]: Parameter key in which urls are stored
        :param extract_zips: bool [optional]: Flag to indicate extraction of the downloaded files
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.param_key: Param[str] = Param(self, "param_key", "")
        self._setDefault(param_key="urls")

        self.download_urls: Param[str] = Param(self, "download_urls", "")
        self._setDefault(download_urls=[])

        self.download_to_path: Param[str] = Param(self, "download_to_path", "")
        self._setDefault(download_to_path="./")

        self.extract_zips: Param[bool] = Param(self, "extract_zips", "")
        self._setDefault(extract_zips=False)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame, response: Dict[str, Any]) -> Dict[str, Any]:  # type: ignore
        param_key: str = self.getParamKey()
        urls: List[str] = self.getDownloadUrls() or response.get(param_key, [])  # type: ignore
        download_path: Optional[str] = self.getDownloadToPath()
        extract_zips: bool = self.getExtractZips()
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        assert urls
        assert download_path

        out: List[str] = []

        for url in urls:
            with ProgressLogMetric(name=name or url, progress_logger=progress_logger):
                downloader = FileDownloader(
                    url=url, download_path=download_path, extract_archives=extract_zips
                )
                result = downloader.download_files_from_url()
                if result:
                    out.append(str(result))

        response["filenames"] = out

        return response

    # noinspection PyPep8Naming
    def getParamKey(self) -> str:
        return cast(str, self.getOrDefault(self.param_key))

    # noinspection PyPep8Naming
    def getDownloadUrls(self) -> str:
        return cast(str, self.getOrDefault(self.download_urls))

    # noinspection PyPep8Naming
    def getDownloadToPath(self) -> Optional[str]:
        return cast(Optional[str], self.getOrDefault(self.download_to_path))

    # noinspection PyPep8Naming
    def getExtractZips(self) -> bool:
        return cast(bool, self.getOrDefault(self.extract_zips))



from collections import UserDict
from typing import Any


class ParameterDict(UserDict):  # type: ignore
    """
    This dictionary returns the value if the key exists otherwise if the key starts with view_
    then it returns the key minus view_
    """

    def __getitem__(self, key: str) -> Any:
        # check if key exists
        if key in super().keys():
            return super().__getitem__(key)
        elif key.startswith("view_"):
            return key[5:]
        else:
            return super().__getitem__(key)

    def copy(self) -> "ParameterDict":
        if self.__class__ is ParameterDict:
            return ParameterDict(self.data.copy())
        import copy

        data = self.data
        try:
            self.data = {}
            c = copy.copy(self)
        finally:
            self.data = data
        c.update(self)
        return c



from pathlib import Path
from typing import List, Union, Dict, Any, Optional, cast

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.utils import AnalysisException

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.file_modes import FileReadModes


class FrameworkParquetLoader(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        view: str,
        file_path: Union[str, List[str], Path],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        merge_schema: bool = False,
        limit: int = -1,
        mode: str = FileReadModes.MODE_PERMISSIVE,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        assert mode in FileReadModes.MODE_CHOICES

        self.logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.file_path: Param[str] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.merge_schema: Param[bool] = Param(self, "merge_schema", "")
        self._setDefault(merge_schema=None)

        self.limit: Param[int] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        path: Union[str, List[str], Path] = self.getFilePath()
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        merge_schema: bool = self.getMergeSchema()
        limit: int = self.getLimit()

        if progress_logger:
            progress_logger.write_to_log(
                f"Loading parquet file for view {view}: {path}"
            )
            progress_logger.log_param(key="data_path", value=str(path))

        with ProgressLogMetric(
            name=f"{name or view}_table_loader", progress_logger=progress_logger
        ):
            try:
                df_reader = df.sql_ctx.read
                df_reader = df_reader.option("mode", self.getMode())

                if merge_schema is True:
                    final_df = (
                        df_reader.option("mergeSchema", "true")
                        .format("parquet")
                        .load(path=str(path))
                    )
                else:
                    final_df = df_reader.format("parquet").load(path=str(path))

                assert (
                    "_corrupt_record" not in final_df.columns
                ), f"Found _corrupt_record after reading the file: {path}. "
                if limit and limit > 0:
                    final_df = final_df.limit(limit)

                # store new data frame in the view
                final_df.createOrReplaceTempView(view)
            except AnalysisException as e:
                self.logger.error(f"File load failed. Location: {path} may be empty")
                raise e
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return cast(str, self.getOrDefault(self.view))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(self) -> Union[str, List[str], Path]:
        return cast(Union[str, List[str], Path], self.getOrDefault(self.file_path))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMergeSchema(self) -> bool:
        return cast(bool, self.getOrDefault(self.merge_schema))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return cast(int, self.getOrDefault(self.limit))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return cast(str, self.getOrDefault(self.mode))



from typing import Any, Dict, Optional, cast

from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class PythonProxyBase(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        verify_count_remains_same: bool = False,
    ) -> None:
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.verify_count_remains_same: Param[bool] = Param(
            self, "verify_count_remains_same", ""
        )

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setName(self, value: str) -> "PythonProxyBase":
        self._paramMap[self.name] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> str:
        return cast(str, self.getOrDefault(self.name or self.__class__.__name__))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setProgressLogger(self, value: ProgressLogger) -> "PythonProxyBase":
        self._paramMap[self.progress_logger] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getProgressLogger(self) -> Optional[ProgressLogger]:
        return cast(ProgressLogger, self.getOrDefault(self.progress_logger))



# type: ignore
from scrapy.crawler import CrawlerProcess


class ScrapyCustomCrawler:
    """
    Implements the crawl process for a scrapy spider
    """

    def __init__(self) -> None:
        """
        Scrapy web crawler for spiders
        """
        self.output = None
        self.process = CrawlerProcess(settings={"LOG_ENABLED": False})

    def yield_output(self, data) -> None:
        """
        Collects yielded data
        :param data: Dict: data to be returned as output
        """
        self.output = data

    def crawl(self, cls: classmethod) -> None:
        """
        Begins the crawl process for a spider
        :param cls: BaseSpiderClass: Implementation of base spider class which has to crawl the web
        """
        self.process.crawl(cls, args={"callback": self.yield_output})
        self.process.start()



from os import getcwd, path
from sys import exit

from spark_pipeline_framework.proxy_generator.proxy_generator import ProxyGenerator


def main() -> int:
    library_folder: str = path.join(getcwd(), "library")
    ProxyGenerator.remove_empty_dirs(library_folder)
    ProxyGenerator.generate_proxies(library_folder)
    return 0


if __name__ == "__main__":
    exit(main())



from typing import Dict, Any

from scrapy.http import Response  # type: ignore

from spark_pipeline_framework.utilities.base_spider_class import BaseSpiderClass


class PythonWgetModuleTestSpider(BaseSpiderClass):
    def __init__(self, **kwargs: Dict[str, Any]):
        super().__init__(
            name="python_wget_module_test",
            urls=["https://pypi.org/project/wget/#files"],
            **kwargs,
        )

    def parse(self, response: Response, **kwargs: Dict[str, Any]) -> Any:
        endpoint = response.xpath('//*[@id="files"]/table/tbody/tr/th/a').attrib["href"]
        self.output.get("response", []).append(endpoint)
        yield self.output



from typing import Any, Dict, Optional, cast

import pymysql
from pymysql import OperationalError
from pymysql.connections import Connection
from pymysql.constants import CLIENT
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkDBQueryRunner(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        username: str,
        password: str,
        host: str,
        port: int,
        query: str,
        db_name: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )
        assert username
        assert password
        assert host
        assert port
        assert query

        self.logger = get_logger(__name__)

        self.username: Param[str] = Param(self, "username", "")
        # noinspection Mypy
        self._setDefault(username=username)

        self.password: Param[str] = Param(self, "password", "")
        # noinspection Mypy
        self._setDefault(password=password)

        self.host: Param[str] = Param(self, "host", "")
        # noinspection Mypy
        self._setDefault(host=host)

        self.port: Param[int] = Param(self, "port", "")
        # noinspection Mypy
        self._setDefault(port=port)

        self.query: Param[str] = Param(self, "query", "")
        # noinspection Mypy
        self._setDefault(query=query)

        self.db_name: Param[Optional[str]] = Param(self, "db_name", "")
        # noinspection Mypy
        self._setDefault(db_name=None)

        # noinspection Mypy
        self._set(**self._input_kwargs)

        super().setStandardParams(
            parameters=parameters, progress_logger=progress_logger
        )

    def _transform(self, df: DataFrame) -> DataFrame:
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        with ProgressLogMetric(name="db_query_runner", progress_logger=progress_logger):
            connection: Connection = pymysql.connect(  # type: ignore
                user=self.getUsername(),
                password=self.getPassword(),
                host=self.getHost(),
                port=self.getPort(),
                db=self.getDb(),
                client_flag=CLIENT.MULTI_STATEMENTS,
            )
            try:
                with connection.cursor() as cursor:
                    rows_affected: int = cursor.execute(self.getQuery())
                    self.logger.info(f"Rows Affected= {rows_affected}")
                connection.commit()

            except OperationalError as e:
                self.logger.error(f"Failed to run query {self.getQuery()}")
                raise e

            finally:
                connection.close()
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getUsername(self) -> str:
        return cast(str, self.getOrDefault(self.username))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getPassword(self) -> str:
        return cast(str, self.getOrDefault(self.password))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getHost(self) -> str:
        return cast(str, self.getOrDefault(self.host))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getPort(self) -> int:
        return cast(int, self.getOrDefault(self.port))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getQuery(self) -> str:
        return cast(str, self.getOrDefault(self.query))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDb(self) -> Optional[str]:
        return cast(Optional[str], self.getOrDefault(self.db_name))



from typing import Dict, Any, Optional, Union, List, Callable

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml import Transformer
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkIfElseTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        enable: Union[bool, Callable[[DataFrame], bool]],
        stages: Union[List[Transformer], Callable[[], List[Transformer]]],
        else_stages: Optional[
            Union[List[Transformer], Callable[[], List[Transformer]]]
        ] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        If enable flag is true then runs stages else runs else_stages
        :param enable: a boolean or a function that takes a DataFrame and returns a boolean
        :param stages: list of transformers or a function that returns a list of transformers
        :param else_stages: list of transformers or a function that returns a list of transformers
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.enable: Union[bool, Callable[[DataFrame], bool]] = enable

        self.stages: Union[List[Transformer], Callable[[], List[Transformer]]] = stages
        self.else_stages: Optional[
            Union[List[Transformer], Callable[[], List[Transformer]]]
        ] = else_stages

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        enable = self.enable if isinstance(self.enable, bool) else self.enable(df)
        if enable:
            stages: List[Transformer] = (
                self.stages if isinstance(self.stages, list) else self.stages()
            )
        else:
            stages = (
                []
                if self.else_stages is None
                else self.else_stages
                if isinstance(self.else_stages, list)
                else self.else_stages()
            )
        for stage in stages:
            if self.progress_logger is not None:
                self.progress_logger.start_mlflow_run(
                    run_name=str(stage), is_nested=True
                )
            df = stage.transform(df)
            if self.progress_logger is not None:
                self.progress_logger.end_mlflow_run()
        return df



from pathlib import Path
from typing import Any, Dict, List

from pyspark.ml.base import Transformer
from pyspark.ml.pipeline import Pipeline
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from library.features.carriers.v1.features_carriers_v1 import FeaturesCarriersV1
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import (
    FrameworkCsvLoader,
)
from spark_pipeline_framework.utilities.pipeline_helper import create_steps


def test_simple_csv_and_sql_pipeline(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    flights_path: str = f"file://{data_dir.joinpath('flights.csv')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    spark_session.sql("DROP TABLE IF EXISTS default.flights")

    # Act
    parameters: Dict[str, Any] = {}

    stages: List[Transformer] = create_steps(
        [
            FrameworkCsvLoader(view="flights", filepath=flights_path),
            FeaturesCarriersV1(parameters=parameters),
        ]
    )

    pipeline: Pipeline = Pipeline(stages=stages)
    transformer = pipeline.fit(df)
    transformer.transform(df)

    # Assert
    result_df: DataFrame = spark_session.sql("SELECT * FROM flights2")
    result_df.show()

    assert result_df.count() > 0



from pathlib import Path
from typing import Union, List, Dict, Any, Optional, cast

from pyspark import keyword_only
from pyspark.ml.param import Param

from pyspark.sql.dataframe import DataFrame

from pyspark.sql.types import StructType

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.transformers.framework_local_file_loader.v1.framework_local_file_loader import (
    FrameworkLocalFileLoader,
)
from spark_pipeline_framework.utilities.file_modes import FileReadModes


class FrameworkJsonLoader(FrameworkLocalFileLoader):
    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    @keyword_only
    def __init__(
        self,
        view: str,
        filepath: Union[str, List[str], Path],
        clean_column_names: bool = False,
        name: Optional[str] = None,
        progress_logger: Optional[ProgressLogger] = None,
        parameters: Optional[Dict[str, Any]] = None,
        limit: int = -1,
        delimiter: str = ",",
        has_header: bool = True,
        infer_schema: bool = False,
        cache_table: bool = True,
        schema: Optional[StructType] = None,
        create_file_path: bool = False,
        use_schema_from_view: Optional[str] = None,
        mode: str = FileReadModes.MODE_PERMISSIVE,
    ) -> None:
        super().__init__(
            view=view,
            filepath=filepath,
            clean_column_names=clean_column_names,
            name=name,
            parameters=parameters,
            progress_logger=progress_logger,
            delimiter=delimiter,
            limit=limit,
            has_header=has_header,
            infer_schema=infer_schema,
            cache_table=cache_table,
            schema=schema,
            create_file_path=create_file_path,
            use_schema_from_view=use_schema_from_view,
            mode=mode,
        )

        self.multiLine: Param[bool] = Param(self, "multiLine", "")
        self._setDefault(multiLine=False)
        self._set(multiLine=False)

    def preprocess(self, df: DataFrame, absolute_paths: List[str]) -> None:
        """
        In pre-processing we try to detect whether the file is a normal json or ndjson
        :param df: DataFrame
        :param absolute_paths: list of paths
        """
        assert absolute_paths
        text_df: DataFrame = df.sql_ctx.read.text(absolute_paths)
        # read the first line of the file
        first_line: str = text_df.select("value").limit(1).collect()[0][0]
        if first_line.lstrip().startswith("["):
            self.setMultiLine(True)
        else:
            self.setMultiLine(False)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setMultiLine(self, value: bool) -> "FrameworkJsonLoader":
        self._paramMap[self.multiLine] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMultiLine(self) -> str:
        return cast(str, self.getOrDefault(self.multiLine))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderFormat(self) -> str:
        return "json"

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getReaderOptions(self) -> Dict[str, Any]:
        options: Dict[str, Any] = {"multiLine": self.getMultiLine()}
        return options



import os
import boto3  # type: ignore
from botocore.exceptions import ClientError  # type: ignore
from smart_open import s3  # type: ignore
from pathlib import Path
from typing import Union, List


def get_absolute_paths(filepath: Union[str, List[str], Path]) -> List[str]:
    """
    Abstracts handling of paths so we can use paths on both k8s in AWS as well as local

    :param filepath: the path or paths to format appropriately
    :return: a list of paths optimized for local or k8s usages
    """
    if not filepath:
        raise ValueError(f"filepath is empty: {filepath}")

    if isinstance(filepath, str):
        if filepath.__contains__(":"):
            return [filepath]
        else:
            data_dir = Path(__file__).parent.parent.joinpath("./")
            return [f"file://{data_dir.joinpath(filepath)}"]
    elif isinstance(filepath, Path):
        data_dir = Path(__file__).parent.parent.joinpath("./")
        return [f"file://{data_dir.joinpath(filepath)}"]
    elif isinstance(filepath, (list, tuple)):
        data_dir = Path(__file__).parent.parent.joinpath("./")
        absolute_paths = []
        for path in filepath:
            if ":" in path:
                absolute_paths.append(path)
            else:
                absolute_paths.append(f"file://{data_dir.joinpath(path)}")
        return absolute_paths
    else:
        raise TypeError(f"Unknown type '{type(filepath)}' for filepath {filepath}")


def isfile(path: str) -> bool:
    if path.startswith("s3://"):
        s3_uri = s3.parse_uri(path)
        bucket = s3_uri["bucket_id"]
        prefix = s3_uri["key_id"]
        s3_resource = boto3.resource("s3")
        try:
            s3_resource.Object(bucket, prefix).load()
            return True
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            else:
                raise e

    return os.path.isfile(path)


def listdir(path: str) -> List[str]:
    query_files = []
    if path.startswith("s3://"):
        s3_uri = s3.parse_uri(path)
        bucket = s3_uri["bucket_id"]
        prefix = s3_uri["key_id"]
        for key, content in s3.iter_bucket(bucket, prefix=prefix):
            query_files.append(f"s3://{bucket}/{key}")
        return query_files
    paths = os.listdir(path)
    for sub_path in paths:
        query_files.append(os.path.join(path, sub_path))
    return query_files



from typing import Any, Dict

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from library.pipelines.reference_data_download_test_pipeline.v1.reference_data_download_test_pipeline import (
    ReferenceDataDownloadTestPipeline,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger


def test_webcrawler_pipeline(spark_session: SparkSession) -> None:
    # Arrange

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    response: Dict[str, Any] = {}

    # Act
    parameters: Dict[str, Any] = {}

    with ProgressLogger() as progress_logger:
        pipeline: ReferenceDataDownloadTestPipeline = ReferenceDataDownloadTestPipeline(
            progress_logger=progress_logger,
            parameters=parameters,
        )
        transformer = pipeline.fit(df, response)
        outcome = transformer.transform(df, response)

    # Assert
    assert len(outcome) > 0



from typing import List, Union, Any

from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable

from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.flattener import flatten


def create_steps(
    my_list: Union[
        List[Transformer],
        List[FrameworkTransformer],
        List[Union[Transformer, List[Transformer]]],
        List[Union[FrameworkTransformer, List[FrameworkTransformer]]],
        List[DefaultParamsReadable[Any]],
    ]
) -> List[Transformer]:
    return flatten(my_list)



class FrameworkMappingRunnerException(Exception):
    def __init__(self, msg: str, exception: Exception) -> None:
        self.exception: Exception = exception
        super().__init__(f"{msg}: \n{exception}")



from pathlib import Path
from typing import Dict, Any, Union, Optional, cast

# noinspection PyProtectedMember
from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.utils import AnalysisException
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class FrameworkJsonExporter(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        file_path: Union[Path, str],
        view: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        limit: int = -1,
        mode: str = FileWriteModes.MODE_OVERWRITE,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert mode in FileWriteModes.MODE_CHOICES

        assert isinstance(file_path, Path) or isinstance(file_path, str)

        assert file_path

        self.logger = get_logger(__name__)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=view)

        self.file_path: Param[Union[Path, str]] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.limit: Param[int] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: Optional[str] = self.getView()
        path: Union[Path, str] = self.getFilePath()
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        # limit: int = self.getLimit()

        with ProgressLogMetric(
            name=f"{name or view}_fhir_exporter", progress_logger=progress_logger
        ):
            try:
                if view:
                    df_view: DataFrame = df.sql_ctx.table(view)
                    assert not spark_is_data_frame_empty(df=df_view)
                    df_view.write.mode(self.getMode()).json(path=str(path))
                else:
                    assert not spark_is_data_frame_empty(df=df)
                    df.write.mode(self.getMode()).json(path=str(path))

                if progress_logger:
                    progress_logger.log_param("data_export_path", str(path))
                    progress_logger.write_to_log(f"[{name or view}] written to {path}")

            except AnalysisException as e:
                self.logger.error(f"[{name or view}]File write failed to {path}")
                raise e
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return cast(Optional[str], self.getOrDefault(self.view))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(self) -> Union[Path, str]:
        return cast(Union[Path, str], self.getOrDefault(self.file_path))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return cast(int, self.getOrDefault(self.limit))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return cast(
            Optional[str], self.getOrDefault(self.name) or self.getOrDefault(self.view)
        )

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return cast(str, self.getOrDefault(self.mode))



from typing import List, Optional, Any, Dict, cast

from pyspark import keyword_only
from pyspark.ml.param import Param
from pyspark.sql import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkDropRowsWithNullTransformer(FrameworkTransformer):
    """
    Drop rows where the column or columns specified in columns_to_check have a null value
    """

    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        view: str,
        columns_to_check: List[str],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__()
        self.logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.columns_to_check: Param[List[str]] = Param(self, "columns_to_check", "")
        self._setDefault(columns_to_check=columns_to_check)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        columns_to_drop: List[str] = self.getColumnsToCheck()
        view: str = self.getView()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        with ProgressLogMetric(
            name=f"{view}_drop_row", progress_logger=progress_logger
        ):
            self.logger.info(
                f"dropping rows if any null values found for columns: {columns_to_drop}"
            )
            df_with_rows: DataFrame = df.sql_ctx.table(view)
            df_with_dropped_rows = df_with_rows.dropna(subset=columns_to_drop)
            df_with_dropped_rows.createOrReplaceTempView(view)
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return cast(str, self.getOrDefault(self.view))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getColumnsToCheck(self) -> List[str]:
        return cast(List[str], self.getOrDefault(self.columns_to_check))



class FileWriteModes:
    MODE_APPEND = "append"
    MODE_OVERWRITE = "overwrite"
    MODE_IGNORE = "ignore"
    MODE_ERROR = "error"
    MODE_CHOICES = (
        MODE_APPEND,
        MODE_OVERWRITE,
        MODE_IGNORE,
        MODE_ERROR,
    )


class FileReadModes:
    MODE_PERMISSIVE = "permissive"
    MODE_DROP_MALFORMED = "dropMalformed"
    MODE_FAIL_FAST = "failFast"
    MODE_CHOICES = (
        MODE_PERMISSIVE,
        MODE_DROP_MALFORMED,
        MODE_FAIL_FAST,
    )



from typing import List, Any, Dict, Union

from scrapy import Spider, Request  # type: ignore
from scrapy.http import Response  # type: ignore


class BaseSpiderClass(Spider):  # type: ignore
    """
    Base class of scrapy spiders.
    All spiders must inherit from this class.
    """

    def __init__(self, name: str, urls: List[str], **kwargs: Any):
        """
        Base Spider Class to crawl the web
        :param name: str: name of the spider
        :param urls: List[str]: list of base urls to be crawled
        """
        super().__init__(name=name, **kwargs)
        self.urls: List[str] = urls
        self.output: Dict[str, Any] = {"response": []}
        self.output_callback = kwargs.get("args", {}).get("callback", None)

    def start_requests(self) -> Any:
        """
        Function to define the start urls and callback methods for a spider
        """
        for url in self.urls:
            yield Request(url=url, callback=self.parse)

    def parse(self, response: Response, **kwargs: Dict[str, Any]) -> Any:
        """
        Function to search for look for elements and yield them.
        """
        self.output = {"response": []}
        yield self.output

    def close(self, spider: Spider, reason: Union[str, Exception]) -> None:
        """
        Function to define tasks to perform before crawl completion
        """
        self.output_callback(self.output)



from typing import Any, Dict, List, Optional, Union

from pyspark.ml.base import Transformer
from pyspark.ml.util import DefaultParamsReadable
from pyspark.sql.dataframe import DataFrame

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.pipelines.framework_pipeline import FrameworkPipeline
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_param_transformer.v1.framework_param_transformer import (
    FrameworkParamTransformer,
)
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.FriendlySparkException import (
    FriendlySparkException,
)
from spark_pipeline_framework.utilities.pipeline_helper import create_steps


class FrameworkParamPipeline(FrameworkPipeline):
    def __init__(
        self, parameters: Dict[str, Any], progress_logger: ProgressLogger
    ) -> None:
        """
        Base class for all file based pipelines
        :param parameters:
        :param progress_logger:
        """
        super(FrameworkParamPipeline, self).__init__(
            parameters=parameters,
            progress_logger=progress_logger,
        )
        self.transformers: List[Union[FrameworkParamTransformer, Transformer]] = []
        self.steps: List[
            Union[
                FrameworkParamTransformer,
                List[FrameworkParamTransformer],
                Transformer,
                List[Transformer],
            ]
        ] = []
        self.__parameters: Dict[str, Any] = parameters
        self.progress_logger: ProgressLogger = progress_logger

    # noinspection PyUnusedLocal
    def fit(self, df: DataFrame, response: Dict[str, Any]) -> "FrameworkParamPipeline":  # type: ignore
        return self

    def _transform(self, df: DataFrame, response: Dict[str, Any]) -> Any:  # type: ignore
        # if steps are defined but not transformers then convert steps to transformers first
        if len(self.steps) > 0 and len(self.transformers) == 0:
            self.transformers = self.create_steps(self.steps)
        # get the logger to use
        logger = get_logger(__name__)
        count_of_transformers: int = len(self.transformers)
        i: int = 0
        pipeline_name: str = self.__class__.__name__
        self.progress_logger.log_event(
            event_name=pipeline_name, event_text=f"Starting Pipeline {pipeline_name}"
        )
        for transformer in self.transformers:
            assert isinstance(transformer, FrameworkTransformer), type(transformer)
            stage_name: Optional[str] = None
            try:
                i += 1
                if hasattr(transformer, "getName"):
                    # noinspection Mypy
                    stage_name = transformer.getName()
                    logger.info(
                        f"---- Running param pipeline [{pipeline_name}] transformer [{stage_name}]  "
                        f"({i} of {count_of_transformers}) ----"
                    )
                else:
                    stage_name = transformer.__class__.__name__
                with ProgressLogMetric(
                    progress_logger=self.progress_logger, name=stage_name or "unknown"
                ):
                    self.progress_logger.log_event(
                        pipeline_name, event_text=f"Running pipeline step {stage_name}"
                    )
                    if isinstance(transformer, FrameworkParamTransformer):
                        response = transformer.transform(dataset=df, response=response)
                    elif isinstance(transformer, FrameworkTransformer):
                        _ = transformer.transform(dataset=df)
            except Exception as e:
                if hasattr(transformer, "getName"):
                    # noinspection Mypy
                    stage_name = transformer.getName()
                else:
                    stage_name = transformer.__class__.__name__
                logger.error(
                    f"!!!!!!!!!!!!! pipeline [{pipeline_name}] transformer [{stage_name}] threw exception !!!!!!!!!!!!!"
                )
                # use exception chaining to add stage name but keep original exception
                friendly_spark_exception: FriendlySparkException = (
                    FriendlySparkException(exception=e, stage_name=stage_name)
                )
                error_messages: List[str] = (
                    friendly_spark_exception.message.split("\n")
                    if friendly_spark_exception.message
                    else []
                )
                for error_message in error_messages:
                    logger.error(msg=error_message)

                if hasattr(transformer, "getSql"):
                    # noinspection Mypy
                    logger.error(transformer.getSql())
                logger.error(
                    "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!"
                )
                self.progress_logger.log_exception(
                    event_name=pipeline_name,
                    event_text=f"Exception in Stage={stage_name}",
                    ex=e,
                )
                raise friendly_spark_exception from e
        self.progress_logger.log_event(
            event_name=pipeline_name, event_text=f"Finished Pipeline {pipeline_name}"
        )
        return response

    def transform(self, dataset: DataFrame, response: Dict[str, Any], params: Dict[Any, Any] = None):  # type: ignore
        if params is None:
            params = dict()
        if isinstance(params, dict):
            if params:
                return self.copy(params)._transform(dataset, response)
            else:
                return self._transform(dataset, response)
        else:
            raise ValueError("Params must be a param map but got %s." % type(params))

    # noinspection PyMethodMayBeStatic
    def create_steps(
        self,
        my_list: Union[
            List[Transformer],
            List[FrameworkTransformer],
            List[FrameworkParamTransformer],
            List[Union[Transformer, List[Transformer]]],
            List[Union[FrameworkTransformer, List[FrameworkTransformer]]],
            List[Union[FrameworkParamTransformer, List[FrameworkParamTransformer]]],
            List[DefaultParamsReadable[Any]],
        ],
    ) -> List[Union[Transformer, FrameworkParamTransformer]]:
        return create_steps(my_list)



import os
from pathlib import Path
from typing import Any, Dict

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType

from spark_pipeline_framework.transformers.framework_file_downloader.v1.framework_file_downloader import (
    FrameworkFileDownloader,
)
from tests.conftest import clean_spark_session


def test_file_downloader(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)
    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    response: Dict[str, Any] = {}

    download_url: str = "https://files.pythonhosted.org/packages/47/6a/62e288da7bcda82b935ff0c6cfe542970f04e29c756b0e147251b2fb251f/wget-3.2.zip"
    download_to_path: str = f"file://{os.path.join(Path(__file__).parent, 'data')}"

    # Act
    response = FrameworkFileDownloader(
        download_urls=[download_url],
        download_to_path=download_to_path,
        extract_zips=True,
    ).transform(df, response)

    # Assert
    assert response



from pathlib import Path

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType

from spark_pipeline_framework.transformers.framework_json_loader.v1.framework_json_loader import (
    FrameworkJsonLoader,
)
from tests.conftest import clean_spark_session


def test_can_load_newline_delimited_json(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_file_path: str = f"{data_dir.joinpath('test.jsonl')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    loader = FrameworkJsonLoader(view="books", filepath=test_file_path, limit=100)
    loader.transform(df)

    # noinspection SqlDialectInspection
    result: DataFrame = spark_session.sql("SELECT * FROM books")

    result.show()

    # Assert
    assert result.count() == 2
    assert loader.getLimit() == 100

    assert result.collect()[1]["title"] == "Essentials of Programming Languages"
    assert len(result.collect()[1]["authors"]) == 2
    assert result.collect()[1]["authors"][0]["surname"] == "Friedman"
    assert result.collect()[1]["edition"] == 3



from pathlib import Path

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType

from spark_pipeline_framework.transformers.framework_json_loader.v1.framework_json_loader import (
    FrameworkJsonLoader,
)
from tests.conftest import clean_spark_session


def test_can_load_simple_json(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)

    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_file_path: str = f"{data_dir.joinpath('test.json')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkJsonLoader(view="books", filepath=test_file_path).transform(df)

    # noinspection SqlDialectInspection
    result: DataFrame = spark_session.sql("SELECT * FROM books")

    result.show()

    # Assert
    assert result.count() == 2

    assert result.collect()[1]["title"] == "Essentials of Programming Languages"
    assert len(result.collect()[1]["authors"]) == 2
    assert result.collect()[1]["authors"][0]["surname"] == "Friedman"
    assert result.collect()[1]["edition"] == 3


def test_can_load_simple_json_with_schema(spark_session: SparkSession) -> None:
    # Arrange
    clean_spark_session(spark_session)
    data_dir: Path = Path(__file__).parent.joinpath("./")
    test_file_path: str = f"{data_dir.joinpath('schema_test.json')}"
    test_file_path_2: str = f"{data_dir.joinpath('schema_test_2.json')}"

    schema = StructType([])

    df: DataFrame = spark_session.createDataFrame(
        spark_session.sparkContext.emptyRDD(), schema
    )

    # Act
    FrameworkJsonLoader(view="books", filepath=test_file_path).transform(df)
    FrameworkJsonLoader(
        view="books_schema", filepath=test_file_path_2, use_schema_from_view="books"
    ).transform(df)
    result: DataFrame = spark_session.sql("SELECT * FROM books")
    result_2: DataFrame = spark_session.sql("SELECT * FROM books_schema")

    # Act
    assert result.schema == result_2.schema



from importlib import import_module
from inspect import signature
from typing import Type, Dict, Any

from pyspark.ml.base import Transformer


class ClassHelpers:
    @staticmethod
    def get_first_class_in_file(full_reference: str) -> Type[Transformer]:
        module = import_module(full_reference)
        md = module.__dict__
        # find the first class in that module (we assume the first class is the Transformer class)
        my_class: Type[Transformer] = [
            md[c]
            for c in md
            if (isinstance(md[c], type) and md[c].__module__ == module.__name__)
        ][0]
        return my_class

    @staticmethod
    def instantiate_class_with_parameters(
        class_parameters: Dict[str, Any], my_class: Type[Any]
    ) -> Any:
        # find the signature of the __init__ method
        my_class_signature = signature(my_class.__init__)
        my_class_args = [
            param.name
            for param in my_class_signature.parameters.values()
            if param.name != "self"
        ]
        # instantiate the class passing in the parameters + progress_logger
        if len(my_class_args) > 0 and len(class_parameters) > 0:
            # noinspection PyArgumentList
            my_instance = my_class(
                **{k: v for k, v in class_parameters.items() if k in my_class_args}
            )
        else:
            my_instance = my_class()
        return my_instance



from pathlib import Path
from typing import Union, Dict, Any, Optional, cast

from pyspark import keyword_only
from pyspark.ml.param import Param

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.transformers.framework_base_exporter.v1.framework_base_exporter import (
    FrameworkBaseExporter,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes


class FrameworkCsvExporter(FrameworkBaseExporter):
    # noinspection PyUnusedLocal
    @keyword_only
    def __init__(
        self,
        file_path: Union[str, Path],
        header: bool,
        delimiter: str = ",",
        view: Optional[str] = None,
        name: Optional[str] = None,
        mode: str = FileWriteModes.MODE_ERROR,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        limit: int = -1,
    ):
        super().__init__(
            view=view,
            name=name,
            mode=mode,
            parameters=parameters,
            progress_logger=progress_logger,
            limit=limit,
        )

        assert isinstance(file_path, Path) or isinstance(file_path, str)
        assert file_path

        self.logger = get_logger(__name__)

        self.file_path: Param[Union[str, Path]] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.header: Param[bool] = Param(self, "header", "")
        self._setDefault(header=header)

        self.delimiter: Param[str] = Param(self, "delimiter", "")
        self._setDefault(delimiter=delimiter)

        self._set(file_path=file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(self) -> Union[str, Path]:
        return cast(Union[str, Path], self.getOrDefault(self.file_path))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getHeader(self) -> bool:
        return cast(bool, self.getOrDefault(self.header))

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDelimiter(self) -> str:
        return cast(str, self.getOrDefault(self.delimiter))

    def getFormat(self) -> str:
        return "csv"

    def getOptions(self) -> Dict[str, Any]:
        return {
            "path": self.getFilePath(),
            "header": self.getHeader(),
            "delimiter": self.getDelimiter(),
        }
